@@ -2,11 +2,11 @@ import json
 import logging
 import requests
 import aiohttp
-import asyncio
 import socket
 import time
 
 from palazzetti_sdk_asset_parser import AssetParser as psap
+from .exceptions import *
 
 _LOGGER = logging.getLogger(__name__)
 
@@ -16,14 +16,14 @@ DISCOVERY_MESSAGE = b"plzbridge?"
 BUFFER_SIZE = 2048
 HTTP_TIMEOUT = 15
 
-#to be completed!!
+# to be completed!!
 class PalComm(object):
     async def async_callUDP(self, host, message):
         server = socket.socket(socket.AF_INET, socket.SOCK_DGRAM, socket.IPPROTO_UDP)
         server.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEPORT, 1)
 
         # Enable broadcasting mode
-        server.setsockopt(socket.SOL_SOCKET, socket.SO_BROADCAST, 1)
+        #server.setsockopt(socket.SOL_SOCKET, socket.SO_BROADCAST, 1)
 
         server.settimeout(DISCOVERY_TIMEOUT)
         server.sendto(message, (host, UDP_PORT))
@@ -86,7 +86,7 @@ class PalComm(object):
 
         # If no response return
         if _response_json["SUCCESS"] != True:
-            #maybe connection error between ConnBox and Stove
+            # maybe connection error between ConnBox and Stove
             return False
 
         return _response_json["DATA"]
@@ -108,25 +108,26 @@ class PalComm(object):
             response = requests.get(queryStr, params=params, timeout=30)
         except requests.exceptions.ReadTimeout:
             # timeout ( can happend when wifi is used )
-            _LOGGER.error('Timeout reach for request : ' + queryStr)
+            _LOGGER.error("Timeout reach for request : " + queryStr)
             return False
         except requests.exceptions.ConnectTimeout:
             # equivalent of ping
-            _LOGGER.error('Please check parm ip : ' + self.ip)
+            _LOGGER.error("Please check parm ip : " + self.ip)
             return False
-        
+
         if response == False:
             return False
-        
+
         _response_json = json.loads(response.text)
-        
+
         # If no response return
         if _response_json["SUCCESS"] != True:
-            #maybe connection error between ConnBox and Stove
+            # maybe connection error between ConnBox and Stove
             return False
 
         return _response_json["DATA"]
 
+
 class PalDiscovery(object):
     async def discovery_UDP(self):
         """discovers all ConnBoxes responding to broadcast"""
@@ -160,26 +161,26 @@ class PalDiscovery(object):
     async def checkIP_UDP(self, testIP):
         """verify the IP is a Connection Box using UDP"""
 
-        api_discovery=PalComm()    
+        api_discovery = PalComm()
         use_ip = testIP
-        
+
         _response = await api_discovery.async_callUDP(use_ip, DISCOVERY_MESSAGE)
-        
+
         if not _response:
             return False
-        
+
         return True
 
     async def checkIP_HTTP(self, testIP):
         """verify the IP is a Connection Box using HTTP call with command GET LABL"""
-        api_discovery=PalComm()    
+        api_discovery = PalComm()
         use_ip = testIP
-        
+
         _response = await api_discovery.async_getHTTP(use_ip, "GET STDT")
-        
+
         if not _response:
             return False
-        
+
         return True
 
     async def checkIP(self, testIP):
@@ -196,11 +197,10 @@ class PalDiscovery(object):
 
         return True
 
+
 class Palazzetti(object):
     """Palazzetti HTTP class"""
 
-    op = None
-
     response_json_alls = None
     response_json_stdt = None
 
@@ -210,12 +210,9 @@ class Palazzetti(object):
     data_config_json = None
     data_config_object = None
 
-    last_op = None
-    last_params = None
-    
     def __init__(self, host, title="uniqueID"):
         self.ip = host
-        self.palsocket=PalComm()
+        self.palsocket = PalComm()
         self.state = "online"
         self.unique_id = title
         self.data = {}
@@ -247,25 +244,50 @@ class Palazzetti(object):
             253: "NOPELLET ALARM",
         }
 
-    # make request GET STDT
+    # TODO: Rename to get_static_data
     async def async_get_stdt(self):
         """Get counters"""
-        self.op = "GET STDT"
         await self.__async_get_request("GET STDT")
 
-    # make request GET ALLS
+    # TODO: Rename to get_alls
     async def async_get_alls(self):
         """Get All data or almost ;)"""
-        self.op = "GET ALLS"
         await self.__async_get_request("GET ALLS")
 
-    # make request GET CNTR
+    async def async_UDP_get_alls(self):
+        """Get All data or almost ;)"""
+        await self.__async_UDP_get_request(b"plzbridge?GET ALLS")
+
+    # TODO: Remove async_ from name
+    async def async_get_label(self):
+        """Get All data or almost ;)"""
+        await self.__async_get_request("GET LABL")
+
+    # TODO: Remove async_ from name
+    async def async_get_status(self):
+        """Get All data or almost ;)"""
+        await self.__async_get_request("GET STAT")
+
+    # TODO: Remove async_ from name
+    async def async_get_fan_data(self):
+        """Get All data or almost ;)"""
+        await self.__async_get_request("GET FAND")
+
+    async def async_get_power(self):
+        """Get All data or almost ;)"""
+        await self.__async_get_request("GET POWR")
+
+    # TODO: Remove async_ from name
+    async def async_get_temperatures(self):
+        """Get All data or almost ;)"""
+        await self.__async_get_request("GET TMPS")
+
+    # TODO: Rename to get_counters
     async def async_get_cntr(self):
         """Get counters"""
-        self.op = "GET CNTR"
         await self.__async_get_request("GET CNTR")
 
-    async def __async_get_request(self,message):
+    async def __async_get_request(self, message):
         """ request the stove """
         _response_json = None
 
@@ -273,45 +295,95 @@ class Palazzetti(object):
         if message is None:
             return False
 
-        # print(self.op)
         _LOGGER.debug("Executing command: {message}")
         # response = False
 
-        #api_discovery=PalComm()
-        #_response = await api_discovery.async_getHTTP(self.ip, message)
+        # api_discovery=PalComm()
+        # _response = await api_discovery.async_getHTTP(self.ip, message)
         _response = await self.palsocket.async_getHTTP(self.ip, message)
 
         if not _response:
             self.state = "offline"
             self.data["state"] = self.state
-            self.response_json.update({"icon": "mdi:link-off"})
+            if self.response_json != None:
+                self.response_json.update({"icon": "mdi:link-off"})
+            else:
+                self.response_json=json.loads('{"icon": "mdi:link-off"}')
             return False
-        
-        #merge the result with the exixting responnse_json
+
+        # merge the result with the exixting responnse_json
         if self.response_json != None:
             response_merged = self.response_json.copy()
             response_merged.update(_response)
             self.response_json = response_merged
         else:
             self.response_json = _response
-        
+
         self.state = "online"
         self.data["state"] = self.state
         self.response_json.update({"icon": "mdi:link"})
         self.data["ip"] = self.ip
         self.__config_parse()
 
-        if self.op == "GET ALLS":
+        if message == "GET ALLS":
             self.data["status"] = self.code_status.get(
                 self.response_json["STATUS"], self.response_json["STATUS"]
             )
             self.response_json_alls = _response
-        elif self.op == "GET STDT":
+        elif message == "GET STDT":
             self.response_json_stdt = _response
+    
+    #only works with GET ALLS
+    async def __async_UDP_get_request(self, message):
+        """ request the stove """
+        _response_json = None
 
+        # check if op is defined or stop here
+        if message is None:
+            return False
+
+        #_LOGGER.debug("Executing command: {message}")
+        _response = await self.palsocket.async_callUDP(self.ip, message)
+        #print(_response)
+        #one single retry
+        if _response is None:
+            #print("retry")
+            _response = await self.palsocket.async_callUDP(self.ip, message)
+
+        if not _response:
+            #print("enter here")
+            self.state = "offline"
+            self.data["state"] = self.state
+            if self.response_json != None:
+                self.response_json.update({"icon": "mdi:link-off"})
+            else:
+                self.response_json=json.loads('{"icon": "mdi:link-off"}')
+            return False
+
+        # merge the result with the exixting responnse_json
+        if self.response_json != None:
+            response_merged = self.response_json.copy()
+            response_merged.update(_response)
+            self.response_json = response_merged
+        else:
+            self.response_json = _response
+
+        self.state = "online"
+        self.data["state"] = self.state
+        self.response_json.update({"icon": "mdi:link"})
+        self.data["ip"] = self.ip
+        self.__config_parse()
+
+        if message == "GET ALLS":
+            self.data["status"] = self.code_status.get(
+                self.response_json["STATUS"], self.response_json["STATUS"]
+            )
+            self.response_json_alls = _response
+        elif message == "GET STDT":
+            self.response_json_stdt = _response
     # send request to stove for set commands
     # why not async?
-    def __request_stove(self, message):
+    def __request_send(self, message):
         """ request the stove """
         _response_json = None
 
@@ -319,13 +391,9 @@ class Palazzetti(object):
         if message is None:
             return False
 
-        # print(self.op)
         _LOGGER.debug("request stove " + message)
         # response = False
 
-        # save
-        self.last_op = message
-
         retry = 0
         success = False
 
@@ -334,8 +402,8 @@ class Palazzetti(object):
         # error returned by Cbox
         while not success:
             # let's go baby
-            #api_discovery=PalComm()
-            #_response = await api_discovery.async_getHTTP(self.ip, message)
+            # api_discovery=PalComm()
+            # _response = await api_discovery.async_getHTTP(self.ip, message)
             _response = self.palsocket.getHTTP(self.ip, message)
 
             # cbox return error
@@ -343,10 +411,16 @@ class Palazzetti(object):
                 # print("palazzetti.stove - com error")
                 self.state = "com error"
                 self.data["state"] = self.state
-                self.response_json.update({"icon": "mdi:link-off"})
+                
+                if self.response_json != None:
+                    self.response_json.update({"icon": "mdi:link-off"})
+                else:
+                    self.response_json=json.loads('{"icon": "mdi:link-off"}')
+
                 _LOGGER.error(
                     "Error returned by CBox - retry in 2 seconds (" + message + ")"
                 )
+                
                 time.sleep(2)
                 retry = retry + 1
 
@@ -357,7 +431,7 @@ class Palazzetti(object):
                         + ")"
                     )
                     break
-            
+
             success = True
 
         # merge the result with the exixting responnse_json
@@ -367,7 +441,7 @@ class Palazzetti(object):
             self.response_json = response_merged
         else:
             self.response_json = _response
-        
+
         self.state = "online"
         self.data["state"] = self.state
         self.response_json.update({"icon": "mdi:link"})
@@ -376,30 +450,110 @@ class Palazzetti(object):
 
         return _response
 
-    # update configuration: call json parse function
-    # async def async_config_parse(self):
-    #     asset_parser = psap(
-    #         get_alls=self.response_json_alls, get_stdt=self.response_json_stdt
-    #     )
-    #     asset_capabilities = asset_parser.parsed_data
-    #     self.data_config_object = asset_capabilities
+    def __validate_power(self, value):
+        _power = None
+
+        try:
+            _power = int(f"{value}", 10)
+        except:
+            _power = None
+
+        if _power == None:
+            raise InvalidPowerError
+
+        if self.data_config_object.flag_has_power == False:
+            raise NotAvailablePowerError
+
+        if _power < 1 or _power > 5:
+            raise InvalidPowerMinMaxError
+
+        return True
+
+    def __validate_fan(self, fan, value):
+        _fan = None
+
+        try:
+            _fan = int(f"{value}", 10)
+        except:
+            _fan = None
+
+        if _fan == None:
+            raise InvalidFanError
+
+        self.async_get_alls()
+
+        _fan_limits = {}
+
+        try:
+            _fan_limits["min"] = self.data_config_object.value_fan_limits[
+                ((fan - 1) * 2)
+            ]
+            _fan_limits["max"] = self.data_config_object.value_fan_limits[
+                (((fan - 1) * 2) + 1)
+            ]
+        except IndexError as error:
+            raise InvalidFanOutOfRange
+        except Exception as error:
+            _fan_limits["min"] = None
+            _fan_limits["max"] = None
+
+        if (_fan_limits.get("min", None) == None) or (_fan_limits.get("max", None)):
+            raise InvalidFanLimitsError
+
+        if _fan < _fan_limits.get("min") or _fan > _fan_limits.get("max"):
+            raise InvalidFanMinMaxError
+
+        return True
+
+    def __validate_setpoint(self, value):
+        _setpoint = None
+
+        try:
+            _setpoint = int(f"{value}", 10)
+        except:
+            _setpoint = None
+
+        if _setpoint == None:
+            raise InvalidSetpointError
+
+        if self.data_config_object.flag_has_setpoint == False:
+            raise NotAvailableSetpointError
+
+        if (
+            _setpoint < self.data_config_object.value_setpoint_min
+            or _setpoint > self.data_config_object.value_setpoint_max
+        ):
+            raise InvalidSetpointMinMaxError
+
+        return True
+
+    def __build_fan_command(self, fan, value):
+
+        _command = {"FAN_1": "SET RFAN", "FAN_2": "SET FN3L", "FAN_3": "SET FN4L"}
+
+        command = f'{_command.get("FAN_" + "" + fan, None)} {value}'
+
+        if command is None:
+            raise InvalidFanOutOfRange
+
+        return command
 
     def __config_parse(self):
-        asset_parser = psap(
-            get_alls=self.response_json, get_stdt=self.response_json
-        )
+        asset_parser = psap(get_alls=self.response_json, get_stdt=self.response_json)
         asset_capabilities = asset_parser.parsed_data
         self.data_config_object = asset_capabilities
 
-    def get_sept(self):
+    def get_setpoint(self):
         """Get target temperature for climate"""
         if self.response_json == None or self.response_json["SETP"] == None:
-            return 
+            return
 
         return self.response_json["SETP"]
 
     # get generic KEY in the datas
     # if key doesn't exist returns None
+    # TODO: Is it possible to remove this function?
+    # BYNOW does not generate risks
     def get_key(self, mykey="STATUS"):
         """Get target temperature for climate"""
         if (
@@ -411,135 +565,164 @@ class Palazzetti(object):
 
         return self.response_json[mykey]
 
-    def set_parameters(self, datas):
+    # TODO: Is it possible to remove this function?
+    # BYNOW it has been limited to setpoint only
+    async def set_parameters(self, datas):
         """set parameters following service call"""
-        self.set_setp(datas.get("SETP", None))  # temperature
-        #self.set_powr(datas.get("PWR", None))  # fire power
-        #self.set_rfan(datas.get("RFAN", None))  # Fan
-        #self.set_status(datas.get("STATUS", None))  # status
+        await self.async_set_setpoint(datas.get("SETP", None))  # temperature
+        # self.set_powr(datas.get("PWR", None))  # fire power
+        # self.set_rfan(datas.get("RFAN", None))  # Fan
+        # self.set_status(datas.get("STATUS", None))  # status
 
-    def set_setp(self, value):
+    async def async_set_label(self, value):
         """Set target temperature"""
-        if value == None or type(value) != int:
-            return
-            
-        if value < self.data_config_object.value_setpoint_min or value > self.data_config_object.value_setpoint_max:
-            return
+        if value == None or value == "":
+            raise InvalidLabelValueError
 
-        if value == self.response_json["SETP"]:
-            return
-        
-        op = "SET SETP"
+        command = f"SET LABL {str(value)}"
 
-        # params for GET
+        if await self.__async_get_request(command) == False:
+            raise SendCommandError
 
-        command = op + " " + str(value)
-        
-        # avoid multiple request
-        if command == self.last_op:
-            _LOGGER.debug("retry for op :" + op + " avoided")
-            return
+        # change state
+        self.data["label"] = value
+        self.response_json.update({"LABEL": value})
 
-        # request the stove
-        if self.__request_stove(command) == False:
-            return
+    async def async_set_fan_silent_mode(self):
+
+        command = self.__build_fan_command(1, 0)
+
+        if self.data_config_object.flag_has_fan_zero_speed_fan == True:
+            command = "SET SLNT 1"
+
+        if await self.__async_get_request(command) == False:
+            raise SendCommandError
+
+    async def async_set_power(self, value):
+
+        self.__validate_power(value)
+
+        command = f"SET POWR {str(value)}"
+
+        if await self.__async_get_request(command) == False:
+            raise SendCommandError
+
+        # change state
+        self.data["powr"] = value
+        self.response_json.update({"POWR": value})
+
+    async def async_set_fan_auto_mode(self, fan=1):
+
+        value = "7"  # Auto Mode
+        command = self.__build_fan_command(fan, value)
+
+        if await self.__async_get_request(command) == False:
+            raise SendCommandError
+
+    async def async_set_fan_high_mode(self, fan=1):
+
+        value = "6"  # High Mode
+        command = self.__build_fan_command(fan, value)
+
+        if await self.__async_get_request(command) == False:
+            raise SendCommandError
+
+    async def async_set_fan(self, fan, value):
+
+        self.__validate_fan(fan, value)
+
+        command = self.__build_fan_command(fan, value)
+
+        if await self.__async_get_request(command) == False:
+            raise SendCommandError
+
+    async def async_set_light(self, value):
+        if (value == None) or (type(value) is not bool):
+            raise InvalidLightError
+
+        command = f"SET LGHT {str(1 if value == True else 0)}"
+
+        if await self.__async_get_request(command) == False:
+            raise SendCommandError
+
+    async def async_set_door(self, value):
+        if (value == None) or (type(value) is not bool):
+            raise InvalidDoorError
+
+        if await self.__async_get_request("GET STAT") == False:
+            raise SendCommandError
+
+        if self.data_config_object.flag_error_status == True:
+            raise InvalidStateError
+
+        command = f"SET DOOR {str(1 if value == True else 2)}"
+
+        if await self.__async_get_request(command) == False:
+            raise SendCommandError
+
+    async def async_set_setpoint(self, value):
+        """Set target temperature"""
+
+        self.__validate_setpoint(value)
+
+        command = f"SET SETP {str(value)}"
+
+        # if value == self.response_json["SETP"]:
+        #     return
+
+        if await self.__async_get_request(command) == False:
+            raise SendCommandError
 
         # change state
         self.data["setp"] = value
         self.response_json.update({"SETP": value})
 
-    def set_powr(self, value):
-        """Set power of fire
-        if value is None :
-            return
+    def power_on(self):
 
-        op = 'SET POWR'
+        if self.__request_send("GET STAT") == False:
+            raise SendCommandError
 
-        # params for GET
-        params = (
-            ('cmd', op + ' ' + str(value)),
-        )
+        if self.data_config_object.flag_error_status == True:
+            raise InvalidStateError
 
-        # avoid multiple request
-        if op == self.last_op and str(params) == self.last_params :
-            _LOGGER.debug('retry for op :' +op+' avoided')
-            return
+        if self.data_config_object.flag_has_switch_on_off != True:
+            raise InvalidStateTransitionError
 
-        # request the stove
-        if self.__request_stove(op, params) == False:
-            return
+        if self.data_config_object.value_product_is_on == False:
+            raise InvalidStateTransitionError
 
-        # change state
-        self.hass.states.set('palazzetti.PWR', self.response_json['PWR'])"""
+        command = "CMD ON"
 
-    def set_rfan(self, value):
-        """Set fan level
+        if self.__request_send(command) == False:
+            raise SendCommandError
 
-        if value == None:
-            return
+    def power_off(self):
 
-        # must be str or int
-        if type(value) != str and type(value) != int:
-            return
+        if self.__request_send("GET STAT") == False:
+            raise SendCommandError
 
-        op = 'SET RFAN'
+        if self.data_config_object.flag_error_status == True:
+            raise InvalidStateError
 
-        # params for GET
-        params = (
-            ('cmd', op + ' ' + str(value)),
-        )
+        if self.data_config_object.flag_has_switch_on_off != True:
+            raise InvalidStateTransitionError
 
-        # avoid multiple request
-        if op == self.last_op and str(params) == self.last_params :
-            _LOGGER.debug('retry for op :' +op+' avoided')
-            return
+        if self.data_config_object.value_product_is_on == False:
+            raise InvalidStateTransitionError
 
-        # request the stove
-        if self.__request_stove(op, params) == False:
-            return
+        command = "CMD OFF"
 
-        # change state
-        self.hass.states.async_set('palazzetti.F2L', self.response_json['F2L'])"""
-
-    def set_status(self, value):
-        """start or stop stove
-
-        fare una GET STAT
-        (questa aggiorna direttamente il config e quindi posso verificare se accendere)
-        controllo il flag_has_switch_onoff
-
-        verificare se già spenta/accesa a seconda dello stato desiderato
-        
-
-        if value == None or type(value) != str :
-            return
-
-        # only ON of OFF value allowed
-        if value != 'on' and value != 'off':
-            return
-
-        op = 'CMD'
-
-        # params for GET
-        params = (
-            ('cmd', op + ' ' + str(value)),
-        )
-
-        # request the stove
-        if self.request_stove(op, params) == False:
-            return
-
-        # change state
-        self.hass.states.async_set('palazzetti.STATUS', self.code_status.get(self.response_json['STATUS'], self.response_json['STATUS']))"""
+        if self.__request_send(command) == False:
+            raise SendCommandError
 
     # retuens list of states: title, state, ip
     def get_data_states(self):
         return self.data
-    
+
     # retuens JSON with all keys of GET ALLS, GET STDT and GET CNTR
     def get_data_json(self):
         return self.response_json
-    
+
     # returns JSON with configuration keys
     def get_data_config_json(self):
         return vars(self.data_config_object)
